# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Utility functions and for loading neurons'''

import glob
import logging
import os
import shutil
import tempfile
import uuid
from functools import partial
from io import IOBase, open

from neurom._compat import StringType, filter
from neurom.core.population import Population
from neurom.exceptions import NeuroMError, RawDataError
from neurom.fst._core import FstNeuron
from neurom.io import neurolucida, swc
from neurom.io.datawrapper import DataWrapper

L = logging.getLogger(__name__)


def _is_morphology_file(filepath):
    """ Check if `filepath` is a file with one of morphology file extensions. """
    return (
        os.path.isfile(filepath) and
        os.path.splitext(filepath)[1].lower() in ('.swc', '.h5', '.asc')
    )


class NeuronLoader(object):
    """
        Caching morphology loader.

        Arguments:
            directory: path to directory with morphology files
            file_ext: file extension to look for (if not set, will pick any of .swc|.h5|.asc)
            cache_size: size of LRU cache (if not set, no caching done)
    """

    def __init__(self, directory, file_ext=None, cache_size=None):
        self.directory = directory
        self.file_ext = file_ext
        if cache_size is not None:
            from pylru import FunctionCacheManager
            self.get = FunctionCacheManager(self.get, size=cache_size)

    def _filepath(self, name):
        """ File path to `name` morphology file. """
        if self.file_ext is None:
            candidates = glob.glob(os.path.join(self.directory, name + ".*"))
            try:
                return next(filter(_is_morphology_file, candidates))
            except StopIteration:
                raise NeuroMError("Can not find morphology file for '%s' " % name)
        else:
            return os.path.join(self.directory, name + self.file_ext)

    # pylint:disable=method-hidden
    def get(self, name):
        """ Get `name` morphology data. """
        return load_neuron(self._filepath(name))


def get_morph_files(directory):
    '''Get a list of all morphology files in a directory

    Returns:
        list with all files with extensions '.swc' , 'h5' or '.asc' (case insensitive)
    '''
    lsdir = (os.path.join(directory, m) for m in os.listdir(directory))
    return list(filter(_is_morphology_file, lsdir))


def get_files_by_path(path):
    '''Get a file or set of files from a file path

    Return list of files with path
    '''
    if os.path.isfile(path):
        return [path]
    elif os.path.isdir(path):
        return get_morph_files(path)

    raise IOError('Invalid data path %s' % path)


def load_neuron(handle, reader=None):
    '''Build section trees from an h5 or swc file'''
    rdw = load_data(handle, reader)
    if isinstance(handle, StringType):
        name = os.path.splitext(os.path.basename(handle))[0]
    else:
        name = None
    return FstNeuron(rdw, name)


def load_neurons(neurons,
                 neuron_loader=load_neuron,
                 name=None,
                 population_class=Population,
                 ignored_exceptions=()):
    '''Create a population object from all morphologies in a directory\
        of from morphologies in a list of file names

    Parameters:
        neurons: directory path or list of neuron file paths
        neuron_loader: function taking a filename and returning a neuron
        population_class: class representing populations
        name (str): optional name of population. By default 'Population' or\
            filepath basename depending on whether neurons is list or\
            directory path respectively.

    Returns:
        neuron population object

    '''
    if isinstance(neurons, (list, tuple)):
        files = neurons
        name = name if name is not None else 'Population'
    elif isinstance(neurons, StringType):
        files = get_files_by_path(neurons)
        name = name if name is not None else os.path.basename(neurons)

    ignored_exceptions = tuple(ignored_exceptions)
    pop = []
    for f in files:
        try:
            pop.append(neuron_loader(f))
        except NeuroMError as e:
            if isinstance(e, ignored_exceptions):
                L.info('Ignoring exception "%s" for file %s',
                       e, os.path.basename(f))
                continue
            raise

    return population_class(pop, name=name)


def _get_file(handle):
    '''Returns the filename of the file to read

    If handle is a stream, a temp file is written on disk first
    and its filename is returned'''
    if not isinstance(handle, IOBase):
        return handle

    fd, temp_file = tempfile.mkstemp(str(uuid.uuid4()), prefix='neurom-')
    os.close(fd)
    with open(temp_file, 'w') as fd:
        handle.seek(0)
        shutil.copyfileobj(handle, fd)
    return temp_file


def load_data(handle, reader=None):
    '''Unpack data into a raw data wrapper'''
    if not reader:
        reader = os.path.splitext(handle)[1][1:].lower()

    if reader not in _READERS:
        raise NeuroMError('Do not have a loader for "%s" extension' % reader)

    filename = _get_file(handle)
    try:
        return _READERS[reader](filename)
    except Exception as e:
        L.exception('Error reading file %s, using "%s" loader', filename, reader)
        raise RawDataError('Error reading file %s:\n%s' % (filename, str(e)))


def _load_h5(filename):
    '''Delay loading of h5py until it is needed'''
    from neurom.io import hdf5
    return hdf5.read(filename,
                     remove_duplicates=False,
                     data_wrapper=DataWrapper)


_READERS = {
    'swc': partial(swc.read,
                   data_wrapper=DataWrapper),
    'h5': _load_h5,
    'asc': partial(neurolucida.read,
                   data_wrapper=DataWrapper)
}



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Generic tree class and iteration functions'''
from collections import deque

from neurom._compat import filter


class Tree(object):
    '''Simple recursive tree class'''
    def __init__(self):
        self.parent = None
        self.children = list()

    def add_child(self, tree):
        '''Add a child to the list of this tree's children

        This tree becomes the added tree's parent
        '''
        tree.parent = self
        self.children.append(tree)
        return tree

    def is_forking_point(self):
        '''Is tree a forking point?'''
        return len(self.children) > 1

    def is_bifurcation_point(self):
        '''Is tree a bifurcation point?'''
        return len(self.children) == 2

    def is_leaf(self):
        '''Is tree a leaf?'''
        return len(self.children) == 0

    def is_root(self):
        '''Is tree the root node?'''
        return self.parent is None

    def ipreorder(self):
        '''Depth-first pre-order iteration of tree nodes'''
        children = deque((self, ))
        while children:
            cur_node = children.pop()
            children.extend(reversed(cur_node.children))
            yield cur_node

    def ipostorder(self):
        '''Depth-first post-order iteration of tree nodes'''
        children = [self, ]
        seen = set()
        while children:
            cur_node = children[-1]
            if cur_node not in seen:
                seen.add(cur_node)
                children.extend(reversed(cur_node.children))
            else:
                children.pop()
                yield cur_node

    def iupstream(self):
        '''Iterate from a tree node to the root nodes'''
        t = self
        while t is not None:
            yield t
            t = t.parent

    def ileaf(self):
        '''Iterator to all leaves of a tree'''
        return filter(Tree.is_leaf, self.ipreorder())

    def iforking_point(self, iter_mode=ipreorder):
        '''Iterator to forking points. Returns a tree object.

        Parameters:
            tree: the tree over which to iterate
            iter_mode: iteration mode. Default: ipreorder.
        '''
        return filter(Tree.is_forking_point, iter_mode(self))

    def ibifurcation_point(self, iter_mode=ipreorder):
        '''Iterator to bifurcation points. Returns a tree object.

        Parameters:
            tree: the tree over which to iterate
            iter_mode: iteration mode. Default: ipreorder.
        '''
        return filter(Tree.is_bifurcation_point, iter_mode(self))

    def __nonzero__(self):
        return bool(self.children)

    __bool__ = __nonzero__



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import numpy as np
import h5py
from neurom.io import hdf5, swc
from neurom.core.dataformat import COLS
from nose import tools as nt


_path = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(_path, '../../../test_data')
H5_PATH = os.path.join(DATA_PATH, 'h5')
H5V1_PATH = os.path.join(H5_PATH, 'v1')
H5V2_PATH = os.path.join(H5_PATH, 'v2')
SWC_PATH = os.path.join(DATA_PATH, 'swc')


def test_get_version():
    with h5py.File(os.path.join(H5V1_PATH, 'Neuron.h5'), mode='r') as v1:
        nt.assert_equal(hdf5.get_version(v1), 'H5V1')

    with h5py.File(os.path.join(H5V2_PATH, 'Neuron.h5'), mode='r') as v2:
        nt.assert_equal(hdf5.get_version(v2), 'H5V2')


def test_unpack_h5():
    with h5py.File(os.path.join(H5V1_PATH, 'Neuron.h5'), mode='r') as v1:
        pts1, grp1 = hdf5._unpack_v1(v1)

    with h5py.File(os.path.join(H5V2_PATH, 'Neuron.h5'), mode='r') as v2:
        pts2, grp2 = hdf5._unpack_v2(v2, stage='raw')

    nt.assert_true(np.all(pts1 == pts2))
    nt.assert_true(np.all(grp1 == grp2))


def test_consistency_between_v1_v2():
    v1_data = hdf5.read(os.path.join(H5V1_PATH, 'Neuron.h5'))
    v2_data = hdf5.read(os.path.join(H5V2_PATH, 'Neuron.h5'))
    nt.ok_(np.allclose(v1_data.data_block, v2_data.data_block))


def test_consistency_between_h5_swc():
    h5_data = hdf5.read(os.path.join(H5V1_PATH, 'Neuron.h5'), remove_duplicates=True)
    swc_data = swc.read(os.path.join(SWC_PATH, 'Neuron.swc'))
    nt.ok_(np.allclose(h5_data.data_block.shape, swc_data.data_block.shape))


class DataWrapper_Neuron(object):
    '''Base class for H5 tests'''

    end_pts = [1, 775, 393, 524, 142, 655, 273, 22, 795, 413,
               544, 162, 675, 293, 423, 42, 815, 564, 182, 695,
               313, 444, 62, 835, 584, 202, 715, 333, 846, 845,
               464, 82, 212, 604, 634, 735, 353, 484, 102, 233,
               624, 755, 373, 504, 122, 253]

    end_parents = [0, 774, 392, 523, 141, 654, 272, 21, 794, 412,
                   543, 161, 674, 292, 422, 41, 814, 563, 181, 694,
                   312, 443, 61, 834, 583, 201, 714, 332, 0, 844,
                   463, 81, 211, 603, 633, 734, 352, 483, 101, 232,
                   623, 754, 372, 503, 121, 252]

    fork_pts = [0, 12, 32, 52, 72, 92, 112, 132, 152, 172, 192, 223,
                243, 263, 283, 303, 323, 343, 363, 383, 403, 434, 454,
                474, 494, 514, 534, 554, 574, 594, 614, 645, 665, 685,
                705, 725, 745, 765, 785, 805, 825]

    fork_parents = [-1, 11, 31, 51, 71, 91, 111, 131, 151, 171, 191, 222,
                    242, 262, 282, 302, 322, 342, 362, 382, 402, 433, 453,
                    473, 493, 513, 533, 553, 573, 593, 613, 644, 664, 684,
                    704, 724, 744, 764, 784, 804, 824]

    def test_n_rows(self):
        nt.assert_equal(self.rows, 927)

    def test_first_id_0(self):
        nt.ok_(self.first_id == 0)


class TestDataWrapper_Neuron_H5V1(DataWrapper_Neuron):
    '''Test HDF5 v1 reading'''
    def setup(self):
        self.data = hdf5.read(os.path.join(H5V1_PATH, 'Neuron.h5'))
        self.first_id = int(self.data.data_block[0][COLS.ID])
        self.rows = len(self.data.data_block)


class TestDataWrapper_Neuron_H5V2(DataWrapper_Neuron):
    '''Test HDF5 v2 reading'''
    def setup(self):
        self.data = hdf5.read(os.path.join(H5V2_PATH, 'Neuron.h5'))
        self.first_id = int(self.data.data_block[0][COLS.ID])
        self.rows = len(self.data.data_block)


class DataWrapper_Neuron_with_duplicates(object):
    '''Base class for H5 tests'''

    end_pts = [1, 386, 133, 782, 529, 914, 276, 661, 23, 408, 155, 925,
               926, 804, 551, 298, 683, 45, 430, 177, 694, 826, 573, 320,
               67, 452, 199, 716, 463, 848, 595, 342, 89, 221, 738, 485,
               870, 232, 617, 364, 111, 760, 507, 892, 254, 639]

    end_parents = [0, 385, 132, 781, 528, 913, 275, 660, 22, 407, 154, 924,
                   0, 803, 550, 297, 682, 44, 429, 176, 693, 825, 572, 319,
                   66, 451, 198, 715, 462, 847, 594, 341, 88, 220, 737, 484,
                   869, 231, 616, 363, 110, 759, 506, 891, 253, 638]

    fork_pts = [0, 12, 34, 56, 78, 100, 122, 144, 166, 188, 210, 243, 265,
                287, 309, 331, 353, 375, 397, 419, 441, 474, 496, 518, 540,
                562, 584, 606, 628, 650, 672, 705, 727, 749, 771, 793, 815,
                837, 859, 881, 903]

    fork_parents = [-1, 11, 33, 55, 77, 99, 121, 143, 165, 187, 209, 242,
                    264, 286, 308, 330, 352, 374, 396, 418, 440, 473, 495,
                    517, 539, 561, 583, 605, 627, 649, 671, 704, 726, 748,
                    770, 792, 814, 836, 858, 880, 902]

    def test_n_rows(self):
        nt.assert_equal(self.rows, 927)

    def test_first_id_0(self):
        nt.ok_(self.first_id == 0)

    def test_end_point_parents(self):
        epar = [self.data.get_parent(i) for i in self.data.get_end_points()]
        nt.assert_equal(epar, DataWrapper_Neuron_with_duplicates.end_parents)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''NeuroM helper utilities'''
from enum import Enum
import json
import warnings
from functools import partial, wraps

import numpy as np


class memoize(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj(object):
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache  # pylint: disable=protected-access
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


def _warn_deprecated(msg):
    '''Issue a deprecation warning'''
    warnings.simplefilter('always', DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    warnings.simplefilter('default', DeprecationWarning)


def deprecated(fun_name=None, msg=""):
    '''Issue a deprecation warning for a function'''
    def _deprecated(fun):
        '''Issue a deprecation warning for a function'''
        @wraps(fun)
        def _wrapper(*args, **kwargs):
            '''Issue deprecation warning and forward arguments to fun'''
            name = fun_name if fun_name is not None else fun.__name__
            _warn_deprecated('Call to deprecated function %s. %s' % (name, msg))
            return fun(*args, **kwargs)

        return _wrapper

    return _deprecated


def deprecated_module(mod_name, msg=""):
    '''Issue a deprecation warning for a module'''
    _warn_deprecated('Module %s is deprecated. %s' % (mod_name, msg))


class NeuromJSON(json.JSONEncoder):
    '''JSON encoder that handles numpy types

    In python3, numpy.dtypes don't serialize to correctly, so a custom converter
    is needed.
    '''

    def default(self, o):  # pylint: disable=method-hidden
        if isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        return json.JSONEncoder.default(self, o)


class OrderedEnum(Enum):
    '''Implementation taken here: https://docs.python.org/3/library/enum.html#orderedenum

    Fixes https://github.com/BlueBrain/NeuroM/issues/697'''

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        raise NotImplementedError

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        raise NotImplementedError

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        raise NotImplementedError

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        raise NotImplementedError



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Distribution configuration for neurom
"""
# pylint: disable=R0801
import os
from setuptools import setup
from setuptools import find_packages


VERSION = "1.4.9"

REQS = ['enum34>=1.0.4;python_version<"3.4"',
        'future>=0.16.0',
        'h5py>=2.7.1',
        'matplotlib>=1.3.1',
        'numpy>=1.8.0',
        'pylru>=1.0',
        'pyyaml>=3.10',
        'scipy>=0.17.0',
        'tqdm>=4.8.4',
        'future>=0.16.0',
        ]

# Hack to avoid installation of modules with C extensions
# in readthedocs documentation building environment.
if os.environ.get('READTHEDOCS') == 'True':
    REQS = ['future>=0.16.0',
            'pyyaml>=3.10',
            ]

config = {
    'description': 'NeuroM: a light-weight neuron morphology analysis package',
    'author': 'BBP Neuroscientific Software Engineering',
    'url': 'http://https://github.com/BlueBrain/NeuroM',
    'version': VERSION,
    'install_requires': REQS,
    'packages': find_packages(),
    'license': 'BSD',
    'scripts': ['apps/raw_data_check',
                'apps/morph_check',
                'apps/morph_stats',
                ],
    'name': 'neurom',
    'extras_require': {
        'plotly': ['plotly==3.0.0'],
    },
    'include_package_data': True,
}

setup(**config)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import numpy as np

from neurom.core.dataformat import COLS
from neurom.io import swc

from nose import tools as nt


_path = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(_path, '../../../test_data')
SWC_PATH = os.path.join(DATA_PATH, 'swc')
SWC_SOMA_PATH = os.path.join(SWC_PATH, 'soma')


def test_read_swc_basic_with_offset_0():
    '''first ID = 0, rare to find'''
    rdw = swc.read(os.path.join(SWC_PATH, 'sequential_trunk_off_0_16pt.swc'))
    nt.eq_(rdw.fmt, 'SWC')
    nt.eq_(len(rdw.data_block), 16)
    nt.eq_(np.shape(rdw.data_block), (16, 7))


def test_read_swc_basic_with_offset_1():
    '''More normal ID numbering, starting at 1'''
    rdw = swc.read(os.path.join(SWC_PATH, 'sequential_trunk_off_1_16pt.swc'))
    nt.eq_(rdw.fmt, 'SWC')
    nt.eq_(len(rdw.data_block), 16)
    nt.eq_(np.shape(rdw.data_block), (16, 7))


def test_read_swc_basic_with_offset_42():
    '''ID numbering starting at 42'''
    rdw = swc.read(os.path.join(SWC_PATH, 'sequential_trunk_off_42_16pt.swc'))
    nt.eq_(rdw.fmt, 'SWC')
    nt.eq_(len(rdw.data_block), 16)
    nt.eq_(np.shape(rdw.data_block), (16, 7))


def test_read_single_neurite():
    rdw = swc.read(os.path.join(SWC_PATH, 'point_soma_single_neurite.swc'))
    nt.eq_(rdw.neurite_root_section_ids(), [1])
    nt.eq_(len(rdw.soma_points()), 1)
    nt.eq_(len(rdw.sections), 2)


def test_read_split_soma():
    rdw = swc.read(os.path.join(SWC_PATH, 'split_soma_single_neurites.swc'))
    nt.eq_(rdw.neurite_root_section_ids(), [1, 3])
    nt.eq_(len(rdw.soma_points()), 3)
    nt.eq_(len(rdw.sections), 4)

    ref_ids = [[-1, 0],
               [0, 1, 2, 3, 4],
               [0, 5, 6],
               [6, 7, 8, 9, 10],
               []]

    for s, r in zip(rdw.sections, ref_ids):
        nt.eq_(s.ids, r)


def test_simple_reversed():
    rdw = swc.read(os.path.join(SWC_PATH, 'simple_reversed.swc'))
    nt.eq_(rdw.neurite_root_section_ids(), [5, 6])
    nt.eq_(len(rdw.soma_points()), 1)
    nt.eq_(len(rdw.sections), 7)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from .utils import get_fig_2d, get_fig_3d  # needs to be at top to trigger matplotlib Agg backend
import os

import itertools as it
import numpy as np

from nose import tools as nt
from neurom import load_neuron
from neurom.view import view, common
from neurom.core import Section
from neurom.core._soma import make_soma, SOMA_CONTOUR, SOMA_CYLINDER
from neurom.core.types import NeuriteType


DATA_PATH = './test_data'
SWC_PATH = os.path.join(DATA_PATH, 'swc/')
fst_neuron = load_neuron(os.path.join(SWC_PATH, 'Neuron.swc'))
simple_neuron = load_neuron(os.path.join(SWC_PATH, 'simple.swc'))


def test_tree():
    with get_fig_2d() as (fig, ax):
        tree = fst_neuron.neurites[0]
        view.plot_tree(ax, tree, color='black', diameter_scale=None, alpha=1., linewidth=1.2)
        collection = ax.collections[0]
        nt.eq_(collection.get_linewidth()[0], 1.2)
        np.testing.assert_allclose(collection.get_color(), np.array([[0., 0., 0., 1.]]))

    with get_fig_2d() as (fig, ax):
        nt.assert_raises(AssertionError, view.plot_tree, ax, tree, plane='wrong')

    with get_fig_2d() as (fig, ax):
        tree = simple_neuron.neurites[0]
        view.plot_tree(ax, tree)
        np.testing.assert_allclose(ax.dataLim.bounds, (-5., 0., 11., 5.), atol=1e-10)


def test_neuron():
    with get_fig_2d() as (fig, ax):
        view.plot_neuron(ax, fst_neuron)
        nt.ok_(ax.get_title() == fst_neuron.name)
        np.testing.assert_allclose(ax.dataLim.get_points(),
                                   [[-40.32853516, -57.600172],
                                    [64.74726272, 48.51626225], ])

    with get_fig_2d() as (fig, ax):
        nt.assert_raises(AssertionError, view.plot_tree, ax, fst_neuron, plane='wrong')


def test_tree3d():
    with get_fig_3d() as (fig, ax):
        tree = simple_neuron.neurites[0]
        view.plot_tree3d(ax, tree)
        xy_bounds = ax.xy_dataLim.bounds
        np.testing.assert_allclose(xy_bounds, (-5., 0., 11., 5.))
        zz_bounds = ax.zz_dataLim.bounds
        np.testing.assert_allclose(zz_bounds, (0., 0., 1., 1.))


def test_neuron3d():
    with get_fig_3d() as (fig, ax):
        view.plot_neuron3d(ax, fst_neuron)
        nt.ok_(ax.get_title() == fst_neuron.name)
        np.testing.assert_allclose(ax.xy_dataLim.get_points(),
                                   [[-40.32853516, -57.600172],
                                    [64.74726272, 48.51626225], ])
        np.testing.assert_allclose(ax.zz_dataLim.get_points().T[0],
                                   (-00.09999862, 54.20408797))


def test_neuron_no_neurites():
    filename = os.path.join(SWC_PATH, 'point_soma.swc')
    with get_fig_2d() as (fig, ax):
        view.plot_neuron(ax, load_neuron(filename))


def test_neuron3d_no_neurites():
    filename = os.path.join(SWC_PATH, 'point_soma.swc')
    with get_fig_3d() as (fig, ax):
        view.plot_neuron3d(ax, load_neuron(filename))


def test_dendrogram():
    with get_fig_2d() as (fig, ax):
        view.plot_dendrogram(ax, fst_neuron)
        np.testing.assert_allclose(ax.get_xlim(), (-20., 100.), rtol=0.25)


def test_one_point_branch():
    test_section = Section(points=np.array([[1., 1., 1., 0.5, 2, 1, 0]]))
    for diameter_scale, linewidth in it.product((1.0, None),
                                                (0.0, 1.2)):
        with get_fig_2d() as (fig, ax):
            view.plot_tree(ax, test_section, diameter_scale=diameter_scale, linewidth=linewidth)
        with get_fig_3d() as (fig, ax):
            view.plot_tree3d(ax, test_section, diameter_scale=diameter_scale, linewidth=linewidth)


soma0 = fst_neuron.soma

# upright, varying radius
soma_2pt_normal_pts = np.array([
    [0.0,   0.0,  0.0, 1.0,  1, 1, -1],
    [0.0,  10.0,  0.0, 10.0, 1, 2,  1],
])
soma_2pt_normal = make_soma(soma_2pt_normal_pts, soma_class=SOMA_CYLINDER)

# upright, uniform radius, multiple cylinders
soma_3pt_normal_pts = np.array([
    [0.0, -10.0,  0.0, 10.0, 1, 1, -1],
    [0.0,   0.0,  0.0, 10.0, 1, 2,  1],
    [0.0,   10.0, 0.0, 10.0, 1, 3,  2],
])
soma_3pt_normal = make_soma(soma_3pt_normal_pts, soma_class=SOMA_CYLINDER)

# increasing radius, multiple cylinders
soma_4pt_normal_pts = np.array([
    [0.0,   0.0,     0.0, 1.0, 1, 1, -1],
    [0.0,   -10.0,   0.0, 2.0, 1, 2, 1],
    [0.0,   -10.0, -10.0, 3.0, 1, 3, 2],
    [-10.0, -10.0, -10.0, 4.0, 1, 4, 3],
])
soma_4pt_normal_cylinder = make_soma(soma_4pt_normal_pts, soma_class=SOMA_CYLINDER)
soma_4pt_normal_contour = make_soma(soma_4pt_normal_pts, soma_class=SOMA_CONTOUR)


def test_soma():
    for s in (soma0,
              soma_2pt_normal,
              soma_3pt_normal,
              soma_4pt_normal_cylinder,
              soma_4pt_normal_contour):
        with get_fig_2d() as (fig, ax):
            view.plot_soma(ax, s)
            common.plt.close(fig)

        with get_fig_2d() as (fig, ax):
            view.plot_soma(ax, s, soma_outline=False)
            common.plt.close(fig)


def test_soma3d():
    with get_fig_3d() as (_, ax):
        view.plot_soma3d(ax, soma_2pt_normal)
        np.testing.assert_allclose(ax.get_xlim(), (-10.,  10.), atol=2)
        np.testing.assert_allclose(ax.get_ylim(), (0.,  10.), atol=2)
        np.testing.assert_allclose(ax.get_zlim(), (-10.,  10.), atol=2)


def test_get_color():
    nt.eq_(view._get_color(None, NeuriteType.basal_dendrite), "red")
    nt.eq_(view._get_color(None, NeuriteType.axon), "blue")
    nt.eq_(view._get_color(None, NeuriteType.apical_dendrite), "purple")
    nt.eq_(view._get_color(None, NeuriteType.soma), "black")
    nt.eq_(view._get_color(None, NeuriteType.undefined), "green")
    nt.eq_(view._get_color(None, 'wrong'), "green")
    nt.eq_(view._get_color('blue', 'wrong'), "blue")
    nt.eq_(view._get_color('yellow', NeuriteType.axon), "yellow")



import os
import numpy as np
from nose import tools as nt
from neurom.core.types import NeuriteType
import neurom.view._dendrogram as dm
from neurom import load_neuron, get

_PWD = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(_PWD, '../../../test_data/h5/v1/Neuron.h5')
NEURON = load_neuron(DATA_PATH)
NEURITE = NEURON.neurites[0]
TREE = NEURITE.root_node

OLD_OFFS = [1.2, -1.2]
NEW_OFFS = [2.3, -2.3]
SPACING = (40., 0.)

def test_n_rectangles_tree():

    nt.assert_equal(dm._n_rectangles(NEURITE), 230)


def test_n_rectangles_neuron():

    nt.assert_equal(dm._n_rectangles(NEURON), 920)


def test_vertical_segment():

    radii = [10., 20.]

    res = np.array([[ -7.7,  -1.2],
                    [-17.7,  -2.3],
                    [ 22.3,  -2.3],
                    [ 12.3,  -1.2]])

    seg = dm._vertical_segment(OLD_OFFS, NEW_OFFS, SPACING, radii)

    nt.assert_true(np.allclose(seg, res))


def test_horizontal_segment():

    diameter = 10.

    res = np.array([[  1.2,  -1.2],
                    [  2.3,  -1.2],
                    [  2.3, -11.2],
                    [  1.2, -11.2]])

    seg = dm._horizontal_segment(OLD_OFFS, NEW_OFFS, SPACING, diameter)

    nt.assert_true(np.allclose(seg, res))


def test_spacingx():

    xoffset = 100.
    xspace = 40.
    max_dims = [10., 2.]

    spx = dm._spacingx(TREE, max_dims, xoffset, xspace)

    nt.assert_almost_equal(spx, -120.)
    nt.assert_almost_equal(max_dims[0], 440.)


def test_update_offsets():

    start_x = -10.
    length = 44.

    offs = dm._update_offsets(start_x, SPACING, 2, OLD_OFFS, length)

    nt.assert_almost_equal(offs[0], 30.)
    nt.assert_almost_equal(offs[1], 42.8)


class TestDendrogram(object):

    def setUp(self):

        self.dtr = dm.Dendrogram(NEURITE)
        self.dnrn = dm.Dendrogram(NEURON)

        self.dtr.generate()
        self.dnrn.generate()

    def test_init(self):

        nt.assert_true(np.allclose(self.dnrn._rectangles.shape, (920, 4, 2)))

    def test_generate_tree(self):

        nt.assert_true(np.allclose(self.dtr._rectangles.shape, (230, 4, 2)))
        nt.assert_false(np.all(self.dtr._rectangles == 0.))

    def test_generate_soma(self):

        vrec = self.dnrn.soma
        trec = np.array([[-0.17071068, -0.34142136],
                         [-0.17071068,  0.        ],
                         [ 0.17071068,  0.        ],
                         [ 0.17071068, -0.34142136]])
        nt.assert_true(np.allclose(vrec, trec))

        vrec = self.dtr.soma

        nt.assert_true(vrec == None)

    def test_neuron_not_corrupted(self):
        # Regression for #492: dendrogram was corrupting
        # neuron used to construct it.
        # This caused the section path distance calculation
        # to raise a KeyError exception.
        get('section_path_distances', NEURON)

    def test_generate_neuron(self):

        total = 0

        for n0, n1 in self.dnrn._groups:

            group = self.dnrn._rectangles[n0: n1]

            total += group.shape[0]

            nt.assert_false(np.all(group == 0.))

        nt.assert_equal(total, 920)

    def test_data(self):

        nt.assert_false(np.all(self.dnrn.data == 0.))
        nt.assert_false(np.all(self.dtr.data == 0.))

    def test_groups(self):
        nt.ok_(self.dnrn.groups)
        nt.ok_(self.dtr.groups)

    def test_dims(self):

        nt.ok_(self.dnrn.dims)
        nt.ok_(self.dtr.dims)

    def test_types_tree(self):
        for ctype in self.dtr.types:
            nt.eq_(ctype, NeuriteType.apical_dendrite)

    def test_types_neuron(self):
        types = tuple(self.dnrn.types)
        nt.eq_(types[0], NeuriteType.apical_dendrite)
        nt.eq_(types[1], NeuriteType.basal_dendrite)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''Module for the detection of the cut plane'''
import logging
import operator

import numpy as np
import neurom as nm

from neurom import viewer
from neurom.core import Tree
from neurom.core.dataformat import COLS


L = logging.getLogger(__name__)


def _create_1d_distributions(points, bin_width):
    '''Create point histograms along each axis

    Parameters:
        points: a np.ndarray of points
        bin_width: the bin width

    Returns: a dict of the X, Y and Z 1D histograms'''
    min_, max_ = np.min(points, axis=0), np.max(points, axis=0)
    hist = dict()
    for i, plane in enumerate('XYZ'):
        # Two binnings with same spacing but different offsets
        # first is attached to min_[i], last is attached to max_[i]
        binning_first_slice = np.arange(min_[i], max_[i] + bin_width, bin_width)
        binning_last_slice = np.arange(max_[i], min_[i] - bin_width, -bin_width)[::-1]
        hist[plane] = (np.histogram(points[:, i], bins=binning_first_slice),
                       np.histogram(points[:, i], bins=binning_last_slice))
    return hist


def _get_probabilities(hist):
    '''Returns -log(p) where p is the a posteriori probabilities of the observed values
    in the bins X min, X max, Y min, Y max, Z min, Z max

    Parameters:
        hist: a dict of the X, Y and Z 1D histograms

    Returns: a dict of -log(p) values'''
    def minus_log_p(mu):
        '''Compute -Log(p) where p is the a posteriori probability to observe 0 counts
        in bin given than the mean value was "mu":
        The number of counts follows a Poisson law so the result is simply... mu
        demo: p(k|mu) = exp(-mu) * mu**k / k!
              p(0|mu) = exp(-mu)
              -log(p) = mu
        '''
        return mu

    for plane, (left, right) in hist.items():
        yield plane, 0, minus_log_p(left[0][0]), left
        yield plane, -1, minus_log_p(right[0][-1]), right


def _get_cut_leaves(neuron, cut_plane_and_position, tolerance):
    '''Returns leaves within cut plane tolerance'''
    cut_plane, position = cut_plane_and_position
    leaves = np.array([leaf.points[-1, COLS.XYZ]
                       for neurite in neuron.neurites
                       for leaf in nm.iter_sections(neurite, iterator_type=Tree.ileaf)])
    idx = 'XYZ'.find(cut_plane)
    return leaves[np.abs(leaves[:, idx] - position) < tolerance]


def draw_neuron(neuron, cut_plane_position, cut_leaves=None):
    '''Draw the neuron in the xy, yz and xz planes.

    Parameters:
        neuron: a Neuron
        cut_plane_position: a tuple (plane, position) like ('Z', 27)
                            that can be specified to add the cut plane on
                            the relevant plots
        cut_leaves: leaves to be highlighted by blue circles
    '''
    figures = dict()
    for draw_plane in ['yz', 'xy', 'xz']:
        fig, axes = viewer.draw(neuron, plane=draw_plane)
        figures[draw_plane] = (fig, axes)

        if cut_leaves is not None:
            import matplotlib
            slice_index = {'yz': [1, 2], 'xy': [0, 1], 'xz': [0, 2]}
            for point in cut_leaves:
                point_2d = point[slice_index[draw_plane]]
                axes.add_artist(matplotlib.patches.Circle(point_2d, radius=2))

        if cut_plane_position:
            cut_plane, position = cut_plane_position
            for axis, line_func in zip(draw_plane, ['axvline', 'axhline']):
                if axis == cut_plane.lower():
                    getattr(axes, line_func)(position)
    return figures


def draw_dist_1d(hist, cut_position=None):
    '''Draw the 1D histograms, optionally also drawing a vertical line at cut_position'''
    import matplotlib.pyplot as plt
    fig = plt.figure()

    h, bins = hist
    plt.hist(bins[:-1], bins=bins, weights=h, label='Point distribution')
    plt.xlabel('Coordinate')
    plt.ylabel('Number of points')

    if cut_position:
        plt.gca().axvline(cut_position, color='r', label='Cut plane')
        plt.legend()
    return {'distrib_1d': (fig, plt.gca())}


def _get_status(minus_log_p):
    '''Returns ok if the probability that there is a cut plane is high enough'''
    _THRESHOLD = 50
    if minus_log_p < _THRESHOLD:
        return 'The probability that there is in fact NO cut plane is high: -log(p) = {0} !'\
            .format(minus_log_p)
    return 'ok'


def find_cut_plane(neuron, bin_width=10, display=False):
    """Find the cut plane

    Parameters:
        neuron: a Neuron object
        bin_width: The size of the binning
        display: where or not to display the control plots
                 Note: It is the user responsability to call matplotlib.pyplot.show()

    Returns:
        A dictionary with the following items:
        status: 'ok' if everything went write, else an informative string
        cut_plane: a tuple (plane, position) where 'plane' is 'X', 'Y' or 'Z'
                   and 'position' is the position
        cut_leaves: an np.array of all termination points in the cut plane
        figures: if 'display' option was used, a dict where values are tuples (fig, ax)
                 for each figure
        details: A dict currently only containing -LogP of the bin where the cut plane was found

    1) The distribution of all points along X, Y and Z is computed
       and put into 3 histograms.

    2) For each histogram we look at the first and last empty bins
       (ie. the last bin before the histogram starts rising,
       and the first after it reaches zero again). Under the assumption
       that there is no cut plane, the posteriori probability
       of observing this empty bin given the value of the not-empty
       neighbour bin is then computed.
    3) The lowest probability of the 6 probabilities (2 for each axes)
       corresponds to the cut plane"""

    points = np.array([point
                       for neurite in (neuron.neurites or [])
                       for section in nm.iter_sections(neurite)
                       for point in section.points])
    if not points.size:
        return {'cut_leaves': None, 'status': "Empty neuron", 'cut_plane': None, 'details': None}

    hist = _create_1d_distributions(points, bin_width)

    cut_plane, side, minus_log_p, histo = max(_get_probabilities(hist), key=operator.itemgetter(2))

    cut_position = histo[1][side]
    cut_leaves = _get_cut_leaves(neuron, (cut_plane, cut_position), bin_width)

    result = {'cut_leaves': cut_leaves,
              'status': _get_status(minus_log_p),
              'details': {'-LogP': minus_log_p},
              'cut_plane': (cut_plane, cut_position)}

    if display:
        result['figures'] = dict()
        result['figures'].update(draw_neuron(neuron, (cut_plane, cut_position), cut_leaves))
        result['figures'].update(draw_dist_1d(histo, cut_position))
        L.info('Trigger the plot display with: matplotlib.pyplot.show()')

    return result
